# -*- coding: utf-8 -*-
"""tantrum utilities module."""
from __future__ import absolute_import
from __future__ import division
from __future__ import print_function
from __future__ import unicode_literals

from . import exceptions
from . import tools
from . import logs
from . import versions

__all__ = ("exceptions", "logs", "tools", "versions")



# -*- coding: utf-8 -*-
"""Python object layer and workflow encapsulation for Tanium's API."""
from __future__ import absolute_import
from __future__ import division
from __future__ import print_function
from __future__ import unicode_literals

from . import adapters
from . import api_clients
from . import api_models
from . import api_objects
from . import auth_methods
from . import exceptions
from . import http_client
from . import results
from . import utils
from . import version
from . import workflows

LOG = utils.logs.LOG
""":obj:`logging.Logger`: Package logger."""

__version__ = version.__version__

__all__ = (
    "adapters",
    "api_clients",
    "api_models",
    "api_objects",
    "auth_methods",
    "exceptions",
    "http_client",
    "results",
    "utils",
    "version",
    "workflows",
)



# -*- coding: utf-8 -*-
"""Python objects for Tanium's API."""
from __future__ import absolute_import
from __future__ import division
from __future__ import print_function
from __future__ import unicode_literals

import six

from .. import api_models
from . import exceptions

float_types = (float,)
""":obj:`tuple` of :obj:`type`: Float types."""

integer_types = six.integer_types
""":obj:`tuple` of :obj:`type`: Integer types."""

string_types = six.string_types
""":obj:`tuple` of :obj:`type`: String types."""

simple_types = tuple(list(integer_types) + list(string_types) + list(float_types))
""":obj:`tuple` of :obj:`type`: All types that should be considered simple types."""

TYPE = "soap"
""":obj:`str`: Type of API this module was built from."""

API_DT = "%Y-%m-%dT%H:%M:%S"
""":obj:`str`: Datetime format for this API type and version."""


def api_fixes():
    """Fix incorrect attribute definitions on auto generated ApiModel classes."""
    models = get_api_all_model()

    fix = {"deleted_flag": "integer_types"}
    models["SavedQuestion"].API_SIMPLE.update(fix)

    fix = {"external_flag": "integer_types"}
    models["User"].API_SIMPLE.update(fix)

    fix = {"cache_row_id": "integer_types"}
    models["AuditData"].API_SIMPLE.update(fix)

    fix = {"cache_info": "CacheInfo"}
    models["AuditDataList"].API_COMPLEX.update(fix)


def expand_cls_globals():
    """Replace str values with global vars for attrs on ApiModel classes.

    Quite the funky chicken dance for solving issues with order of class definition.
    """
    item_classes = ApiItem.__subclasses__()
    list_classes = ApiList.__subclasses__()
    for c in item_classes:
        c.API_SIMPLE = expand_global(obj=c.API_SIMPLE)
        c.API_COMPLEX = expand_global(obj=c.API_COMPLEX)
        c.API_LIST_CLS = expand_global(obj=c.API_LIST_CLS)
    for c in list_classes:
        c.API_SIMPLE = expand_global(obj=c.API_SIMPLE)
        c.API_COMPLEX = expand_global(obj=c.API_COMPLEX)
        c.API_ITEM_CLS = expand_global(obj=c.API_ITEM_CLS)


def expand_global(obj):
    """Replace str values with global vars recursively.

    Args:
        obj (:obj:`str` or :obj:`dict` or :obj:`list`):
            Object to expand into global var.

    Returns:
        :obj:`str` or :obj:`dict` or :obj:`list` or :obj:`object`:
            String objects are the only ones that will be expanded.

    """
    if isinstance(obj, string_types):
        if obj == "None":
            return None
        try:
            return globals()[obj]
        except Exception:
            return obj
    if isinstance(obj, dict):
        return {k: expand_global(obj=v) for k, v in obj.items()}
    if isinstance(obj, (list, tuple)):
        return [expand_global(obj=v) for v in obj]
    return obj


def get_api_all_item():
    """Get map of class name to class for all ApiItem subclasses.

    Returns:
        :obj:`dict`

    """
    return {c.__name__: c for c in ApiItem.__subclasses__()}


def get_api_all_list():
    """Get map of class name to class for all ApiList subclasses.

    Returns:
        :obj:`dict`

    """
    return {c.__name__: c for c in ApiList.__subclasses__()}


def get_api_all_model():
    """Get map of class name to class for all ApiItem and ApiList subclasses.

    Returns:
        :obj:`dict`

    """
    return {c.__name__: c for c in ApiItem.__subclasses__() + ApiList.__subclasses__()}


class ApiModel(api_models.ApiModel):
    """Model for a complex item in the API."""

    pass


class ApiItem(ApiModel, api_models.ApiItem):
    """Model for a complex item in the API."""

    pass


class ApiList(ApiModel, api_models.ApiList):
    """Model for an array in the API."""

    pass


class ResultInfoList(ApiList):
    """Manually defined API array object."""

    API_NAME = "result_infos"
    """:obj:`str`: Name of object used in API calls."""

    API_NAME_SRC = ""
    """:obj:`str`: Name of object in source file."""

    API_SIMPLE = {"now": "string_types", "max_available_age": "integer_types"}
    """:obj:`dict`: Map of simple attributes to their types."""

    API_COMPLEX = {}
    """:obj:`dict`: Map of complex attributes to their types."""

    API_CONSTANTS = {}
    """:obj:`dict`: Map of constants to their values."""

    API_STR_ADD = ["now", "max_available_age"]
    """:obj:`list` or :obj:`str`: Attributes to add to str formatting."""

    API_ITEM_ATTR = "result_info"
    """:obj:`str`: Name of :attr:`API_ITEM_CLS` used in API calls."""

    API_ITEM_CLS = "ResultInfo"
    """:class:`ApiItem`: Item class this list class holds."""

    def __repr__(self):
        """Make repr output be same as str output.

        Returns:
            :obj:`str`

        """
        return self.__str__()


class ResultInfo(ApiItem):
    """Manually defined API object."""

    API_NAME = "result_info"
    """:obj:`str`: Name of object used in API calls."""

    API_NAME_SRC = ""
    """:obj:`str`: Name of object in source file."""

    API_SIMPLE = {
        "age": "integer_types",
        "archived_question_id": "integer_types",
        "error_count": "integer_types",
        "estimated_total": "integer_types",
        "expire_seconds": "integer_types",
        "id": "integer_types",
        "issue_seconds": "integer_types",
        "mr_passed": "integer_types",
        "mr_tested": "integer_types",
        "no_results_count": "integer_types",
        "passed": "integer_types",
        "question_id": "integer_types",
        "report_count": "integer_types",
        "row_count": "integer_types",
        "row_count_flag": "integer_types",
        "row_count_machines": "integer_types",
        "saved_question_id": "integer_types",
        "seconds_since_issued": "integer_types",
        "select_count": "integer_types",
        "tested": "integer_types",
    }
    """:obj:`dict`: Map of simple attributes to their types."""

    API_COMPLEX = {}
    """:obj:`dict`: Map of complex attributes to their types."""

    API_CONSTANTS = {}
    """:obj:`dict`: Map of constants to their values."""

    API_STR_ADD = [
        "question_id",
        "estimated_total",
        "row_count",
        "mr_tested",
        "mr_passed",
    ]
    """:obj:`list` or :obj:`str`: Attributes to add to str formatting."""

    API_LIST_CLS = "ResultInfoList"
    """:class:`ApiList`: List class that holds this item class."""

    API_LIST_API_NAME = "result_infos"
    """:obj:`str`: Name of :attr:`API_LIST_CLS` used in API calls."""

    def __repr__(self):
        """Make repr output be same as str output.

        Returns:
            :obj:`str`

        """
        return self.__str__()


class ResultSetList(ApiList):
    """Manually defined API array object."""

    API_NAME = "result_sets"
    """:obj:`str`: Name of object used in API calls."""

    API_NAME_SRC = ""
    """:obj:`str`: Name of object in source file."""

    API_SIMPLE = {"now": "string_types"}
    """:obj:`dict`: Map of simple attributes to their types."""

    API_COMPLEX = {}
    """:obj:`dict`: Map of complex attributes to their types."""

    API_CONSTANTS = {}
    """:obj:`dict`: Map of constants to their values."""

    API_STR_ADD = ["now"]
    """:obj:`list` or :obj:`str`: Attributes to add to str formatting."""

    API_ITEM_ATTR = "result_set"
    """:obj:`str`: Name of :attr:`API_ITEM_CLS` used in API calls."""

    API_ITEM_CLS = "ResultSet"
    """:class:`ApiItem`: Item class this list class holds."""


class MergedResultSet(ApiItem):
    """Manually defined API object."""

    API_NAME = "merged_result"
    """:obj:`str`: Name of object used in API calls."""

    API_NAME_SRC = ""
    """:obj:`str`: Name of object in source file."""

    API_SIMPLE = {"now": "string_types", "max_available_age": "string_types"}
    """:obj:`dict`: Map of simple attributes to their types."""

    API_COMPLEX = {"result_infos": "ResultInfoList", "result_sets": "ResultSetList"}
    """:obj:`dict`: Map of complex attributes to their types."""

    API_CONSTANTS = {}
    """:obj:`dict`: Map of constants to their values."""

    API_STR_ADD = ["now", "max_available_age"]
    """:obj:`list` or :obj:`str`: Attributes to add to str formatting."""

    API_LIST_CLS = None
    """:class:`ApiList`: List class that holds this item class."""

    API_LIST_API_NAME = ""
    """:obj:`str`: Name of :attr:`API_LIST_CLS` used in API calls."""


class ResultSet(ApiItem):
    """Manually defined API object."""

    API_NAME = "result_set"
    """:obj:`str`: Name of object used in API calls."""

    API_NAME_SRC = ""
    """:obj:`str`: Name of object in source file."""

    API_SIMPLE = {
        "age": "integer_types",
        "archived_question_id": "integer_types",
        "cache_id": "integer_types",
        "error_count": "integer_types",
        "estimated_total": "integer_types",
        "expiration": "integer_types",
        "expire_seconds": "integer_types",
        "filtered_row_count": "integer_types",
        "filtered_row_count_machines": "integer_types",
        "id": "integer_types",
        "issue_seconds": "integer_types",
        "item_count": "integer_types",
        "mr_passed": "integer_types",
        "mr_tested": "integer_types",
        "no_results_count": "integer_types",
        "passed": "integer_types",
        "question_id": "integer_types",
        "report_count": "integer_types",
        "row_count": "integer_types",
        "row_count_machines": "integer_types",
        "saved_question_id": "integer_types",
        "seconds_since_issued": "integer_types",
        "select_count": "integer_types",
        "tested": "integer_types",
    }
    """:obj:`dict`: Map of simple attributes to their types."""

    API_COMPLEX = {"cs": "ColumnList", "rs": "RowList"}
    """:obj:`dict`: Map of complex attributes to their types."""

    API_CONSTANTS = {}
    """:obj:`dict`: Map of constants to their values."""

    API_STR_ADD = [
        "question_id",
        "estimated_total",
        "mr_tested",
        "mr_passed",
        "row_count",
        "columns",
    ]
    """:obj:`list` or :obj:`str`: Attributes to add to str formatting."""

    API_LIST_CLS = "ResultSetList"
    """:class:`ApiList`: List class that holds this item class."""

    API_LIST_API_NAME = "result_sets"
    """:obj:`str`: Name of :attr:`API_LIST_CLS` used in API calls."""

    def __init__(self, **kwargs):
        """Constructor.

        Notes:
            Sets :attr:`Column.API_DATA_SET` so column objects can access rows.

            Sets :attr:`RowColumn.API_COLUMN` to the index correlated :obj:`Column`
            so row columns can access the column name, result type, etc.

            Sets :attr:`Column.API_IDX` and :attr:`RowColumn.API_IDX` so column
            objects know what their index is without having to do lookups.

        Args:
            **kwargs:
                Passed back to parent class.

        """
        super(ResultSet, self).__init__(**kwargs)
        self.cs = ColumnList() if self.cs is None else self.cs
        self.rs = RowList() if self.rs is None else self.rs
        for row in self.rows:
            for idx, row_column in enumerate(row):
                row_column.API_COLUMN = self.columns[idx]
                row_column.API_IDX = idx
        for idx, column in enumerate(self.columns):
            column.API_DATA_SET = self
            column.API_IDX = idx

    @property
    def rows(self):
        """Expose complex attr "rs" as "rows".

        Returns:
            :obj:`list`

        """
        return self.rs

    @property
    def columns(self):
        """Expose complex attr "cs" as "columns".

        Returns:
            :obj:`list`

        """
        return self.cs


class ColumnList(ApiList):
    """Manually defined API array object."""

    API_NAME = "cs"
    """:obj:`str`: Name of object used in API calls."""

    API_NAME_SRC = ""
    """:obj:`str`: Name of object in source file."""

    API_SIMPLE = {}
    """:obj:`dict`: Map of simple attributes to their types."""

    API_COMPLEX = {}
    """:obj:`dict`: Map of complex attributes to their types."""

    API_CONSTANTS = {}
    """:obj:`dict`: Map of constants to their values."""

    API_STR_ADD = ["names"]
    """:obj:`list` or :obj:`str`: Attributes to add to str formatting."""

    API_ITEM_ATTR = "c"
    """:obj:`str`: Name of :attr:`API_ITEM_CLS` used in API calls."""

    API_ITEM_CLS = "Column"
    """:class:`ApiItem`: Item class this list class holds."""

    def __repr__(self):
        """Make repr output be same as str output.

        Returns:
            :obj:`str`

        """
        return self.__str__()

    @property
    def names(self):
        """Expose column names of all :obj:`Column` in this obj.

        Returns:
            :obj:`list` of :obj:`str`

        """
        return [x.name for x in self]

    def __getitem__(self, value):
        """Support indexing of :attr:`names`.

        Args:
            value (:obj:`int` or :obj:`str`):
                If int: Index of item to retrieve from simple attr "c".
                If str: Column name to index correlate from :attr:`names`

        Returns:
            :attr:`RowColumn`:
                The RowColumn object at index value from :attr:`names`.

        """
        if isinstance(value, string_types):
            try:
                value = self.names.index(value)
            except ValueError:
                error = [
                    "Valid Column Names:",
                    "{vc}",
                    "{c!r} is an invalid column name in this data set",
                ]
                error = "\n  ".join(error)
                error = error.format(c=value, vc=self.names)
                raise exceptions.ModuleError(error)
        return self.LIST[value]


class Column(ApiItem):
    """Manually defined API object."""

    API_DATA_SET = None
    """:obj:`ResultSet`: Parent of this object, set by :meth:`ResultSet.__init__`."""

    API_IDX = None
    """:obj:`int`: Index of this object in :obj:`ColumnList`."""

    API_NAME = "c"
    """:obj:`str`: Name of object used in API calls."""

    API_NAME_SRC = ""
    """:obj:`str`: Name of object in source file."""

    API_SIMPLE = {"dn": "string_types", "rt": "integer_types", "wh": "integer_types"}
    """:obj:`dict`: Map of simple attributes to their types."""

    API_COMPLEX = {}
    """:obj:`dict`: Map of complex attributes to their types."""

    API_CONSTANTS = {}
    """:obj:`dict`: Map of constants to their values."""

    API_STR_ADD = ["name", "result_type", "hash"]
    """:obj:`list` or :obj:`str`: Attributes to add to str formatting."""

    API_LIST_CLS = "ColumnList"
    """:class:`ApiList`: List class that holds this item class."""

    API_LIST_API_NAME = "cs"
    """:obj:`str`: Name of :attr:`API_LIST_CLS` used in API calls."""

    @property
    def name(self):
        """Expose simple attr "dn" as name.

        Returns:
            :obj:`str`

        """
        return self.dn

    @property
    def hash(self):
        """Expose simple attr "wh" as hash.

        Returns:
            :obj:`str`

        """
        return self.wh

    @property
    def type(self):
        """Expose simple attr "rt" as type.

        Returns:
            :obj:`str`

        """
        return self.rt

    @property
    def result_type(self):
        """Expose simple attr "rt" as result_type.

        Notes:
            Will try to map str from "rt" as int to constants from
            :attr:`Sensor.API_CONSTANTS`.

        Returns:
            :obj:`str`

        """
        sensor = get_api_all_model()["Sensor"]
        constants = sensor.API_CONSTANTS.items()
        result_map = {v: k for k, v in constants if not k.lower().endswith("sensor")}
        result_type = self.api_coerce_int(self.type)
        return result_map.get(result_type, result_type)

    def get_values(self, attr="value", join=None):
        """Get values of this column from all rows.

        Args:
            attr (:obj:`str`, optional):
                Attribute to get values of.

                Defaults to: "value".
            join (:obj:`str`, optional):
                String to join multi value columns.

                Defaults to: None.

        Returns:
            :obj:`list` of :obj:`str`

        """
        values = []
        for row in self.API_DATA_SET.rows:
            row_column = row[self.API_IDX]
            row_column_values = row_column.get_values(attr=attr, join=join)
            values.append(row_column_values)
        return values


class RowList(ApiList):
    """Manually defined API array object."""

    API_NAME = "rs"
    """:obj:`str`: Name of object used in API calls."""

    API_NAME_SRC = ""
    """:obj:`str`: Name of object in source file."""

    API_SIMPLE = {}
    """:obj:`dict`: Map of simple attributes to their types."""

    API_COMPLEX = {}
    """:obj:`dict`: Map of complex attributes to their types."""

    API_CONSTANTS = {}
    """:obj:`dict`: Map of constants to their values."""

    API_STR_ADD = []
    """:obj:`list` or :obj:`str`: Attributes to add to str formatting."""

    API_ITEM_ATTR = "r"
    """:obj:`str`: Name of :attr:`API_ITEM_CLS` used in API calls."""

    API_ITEM_CLS = "Row"
    """:class:`ApiItem`: Item class this list class holds."""

    def __repr__(self):
        """Make repr output be same as str output.

        Returns:
            :obj:`str`

        """
        return self.__str__()


class Row(ApiItem):
    """Manually defined API object."""

    API_NAME = "r"
    """:obj:`str`: Name of object used in API calls."""

    API_NAME_SRC = ""
    """:obj:`str`: Name of object in source file."""

    API_SIMPLE = {"id": "integer_types", "cid": "integer_types"}
    """:obj:`dict`: Map of simple attributes to their types."""

    API_COMPLEX = {"c": "RowColumnList"}
    """:obj:`dict`: Map of complex attributes to their types."""

    API_CONSTANTS = {}
    """:obj:`dict`: Map of constants to their values."""

    API_STR_ADD = ["id", "cid", "columns"]
    """:obj:`list` or :obj:`str`: Attributes to add to str formatting."""

    API_LIST_API_NAME = "rs"
    """:obj:`str`: Name of :attr:`API_LIST_CLS` used in API calls."""

    API_LIST_CLS = "RowList"
    """:class:`ApiItem`: Item class this list class holds."""

    def __repr__(self):
        """Make repr output be same as str output.

        Returns:
            :obj:`str`

        """
        return self.__str__()

    def __getitem__(self, value):
        """Support indexing of :attr:`names`.

        Args:
            value (:obj:`int` or :obj:`str`):
                If int: Index of item to retrieve from :attr:`names`.
                If str: Column name to index correlate from :attr:`ColumnList.names`

        Returns:
            :attr:`RowColumn`:
                The RowColumn object at index value from :attr:`names`.

        """
        if isinstance(value, string_types):
            try:
                value = self.names.index(value)
            except ValueError:
                error = [
                    "Valid Column Names:",
                    "{vc}",
                    "{c!r} is an invalid column name in this data set",
                ]
                error = "\n  ".join(error)
                error = error.format(c=value, vc=self.names)
                raise exceptions.ModuleError(error)
        return self.columns[value]

    @property
    def names(self):
        """Return the names of each row column from :attr:`RowColumn.API_COLUMN`.

        Returns:
            :obj:`list` of :obj:`str`

        """
        return [x.API_COLUMN.name for x in self.columns]

    @property
    def columns(self):
        """Expose complex attr "c" as "columns".

        Returns:
            :obj:`RowColumnList`

        """
        return self.c

    def __len__(self):
        """Expose length of :attr:`columns`.

        Returns:
            :obj:`int`

        """
        return len(self.columns)

    def get_values(self, meta=False, hashes=False, join=None):
        """Get all values from all columns in this row.

        Args:
            meta (:obj:`bool`, optional):
                Include sensor hash and result type for each column.

                Defaults to: False.

            hashes (:obj:`bool`, optional):
                Include hashes of values.

                Defaults to: False.
            join (:obj:`str`, optional):
                String to join multi value columns.

                Defaults to: None.

        Returns:
            :obj:`dict`

        """
        values = {}
        for idx, column in enumerate(self.columns):
            values[column.name] = column.get_values(attr="value", join=join)
            if hashes:
                key = "{} Hash Values".format(column.name)
                values[key] = column.get_values(attr="hash", join=join)
            if meta:
                key = "{} Sensor Hash".format(column.name)
                values[key] = column.hash
                key = "{} Result Type".format(column.name)
                values[key] = column.result_type
        return values


class RowColumnList(ApiList):
    """Manually defined API array object."""

    API_NAME = "rc"
    """:obj:`str`: Name of object used in API calls."""

    API_NAME_SRC = ""
    """:obj:`str`: Name of object in source file."""

    API_SIMPLE = {}
    """:obj:`dict`: Map of simple attributes to their types."""

    API_COMPLEX = {}
    """:obj:`dict`: Map of complex attributes to their types."""

    API_CONSTANTS = {}
    """:obj:`dict`: Map of constants to their values."""

    API_STR_ADD = []
    """:obj:`list` or :obj:`str`: Attributes to add to str formatting."""

    API_ITEM_ATTR = "row_columns"
    """:obj:`str`: Name of :attr:`API_ITEM_CLS` used in API calls."""

    API_ITEM_CLS = "RowColumn"
    """:class:`ApiItem`: Item class this list class holds."""

    def api_coerce_items_hook(self, attr, value, op):
        """Check hook that allows subclasses to modify list items.

        Args:
            attr (:obj:`str`):
                Attribute being set on this object.
            value (:obj:`ApiList` or :obj:`list` or :obj:`tuple`):
                The list object holding items.
            op (:obj:`str`):
                Operation being performed that called this method.

        Examples:
            This converts all "v" attributes of each column for this row into a
            list of dicts:

                >>> # single value columns "v" is a single dict:
                >>> {"v": {"h": "", "text": ""}}
                >>> # multiple value columns "v" is a list of dict:
                >>> {"v": [{"h": "", "text": ""}, {"h": "", "text": ""}]}
                >>> # Count columns "v" is a string:
                >>> {"v": "1"}
                >>> # all become:
                >>> {"v": [{"text": "", "hash": ""}]}

        Notes:
            The driving force behind this is basically just the Count column.
            It returns a string of the value, instead of a dict or a list of dicts.

        Returns:
            :obj:`list`:
                The list of items modified or as is.

        """
        if not isinstance(value, (list, tuple)):
            error = "{value} is not a list"
            error = error.format(value=value)
            raise exceptions.ModuleError(error)

        new_value = []
        for item in value:
            if not isinstance(item, dict):
                error = "Item {item} in {value} is not a dict"
                error = error.format(item=item, value=value)
                raise exceptions.ModuleError(error)

            item_v = item["v"]
            if isinstance(item_v, (tuple, list)):
                new_item_value = []
                for item_value in item_v:
                    is_simple = isinstance(item_value, simple_types)
                    is_none = item_value is None
                    item_value = (
                        {"text": item_value, "h": None} if is_simple else item_value
                    )

                    item_value = {"text": "", "h": None} if is_none else item_value

                    if not isinstance(item_value, dict):
                        error = "Item value '{item_value}' in {item} is not a dict"
                        error = error.format(item_value=item_value, item=item["v"])
                        raise exceptions.ModuleError(error)

                    new_item_value.append(item_value)
                item_v = new_item_value

            is_none = item_v is None
            is_simple = isinstance(item_v, simple_types)
            item_v = [{"text": "", "h": None}] if is_none else item_v
            item_v = [{"text": item["v"], "h": None}] if is_simple else item_v
            item_v = self.api_coerce_list(item_v)
            item["v"] = item_v
            new_value.append(item)
        return new_value


'''
# older version of hook, it didn't work right :)
    def api_coerce_items_hook(self, attr, value, op):
        """Check hook that allows subclasses to modify list items.

        Args:
            attr (:obj:`str`):
                Attribute being set on this object.
            value (:obj:`ApiList` or :obj:`list` or :obj:`tuple`):
                The list object holding items.
            op (:obj:`str`):
                Operation being performed that called this method.

        Examples:
            This converts all "v" attributes of each column for this row into a
            list of dicts:

                >>> # single value columns "v" is a single dict:
                >>> {"v": {"h": "", "text": ""}}
                >>> # multiple value columns "v" is a list of dict:
                >>> {"v": [{"h": "", "text": ""}, {"h": "", "text": ""}]}
                >>> # Count columns "v" is a string:
                >>> {"v": "1"}
                >>> # all become:
                >>> {"v": [{"text": "", "hash": ""}]}

        Notes:
            The driving force behind this is basically just the Count column.
            It returns a string of the value, instead of a dict or a list of dicts.

        Returns:
            :obj:`list`:
                The list of items modified or as is.

        """
        items = []
        for item in value:
            do_fix = isinstance(item, dict) and isinstance(item["v"], string_types)
            item["v"] = {"text": item["v"], "h": None} if do_fix else item["v"]
            item["v"] = self.api_coerce_list(item["v"])
            items.append(item)
        return items
'''


class RowColumn(ApiList):
    """Manually defined API array object."""

    API_COLUMN = None
    """:obj:`Column`: Index correlated column, set by :meth:`ResultSet.__init__`."""

    API_IDX = None
    """:obj:`int`: Index of this object in :obj:`RowColumnList`."""

    API_NAME = "row_column"
    """:obj:`str`: Name of object used in API calls."""

    API_NAME_SRC = ""
    """:obj:`str`: Name of object in source file."""

    API_SIMPLE = {}
    """:obj:`dict`: Map of simple attributes to their types."""

    API_COMPLEX = {}
    """:obj:`dict`: Map of complex attributes to their types."""

    API_CONSTANTS = {}
    """:obj:`dict`: Map of constants to their values."""

    API_STR_ADD = ["name", "hash", "result_type"]
    """:obj:`list` or :obj:`str`: Attributes to add to str formatting."""

    API_ITEM_ATTR = "v"
    """:obj:`str`: Name of :attr:`API_ITEM_CLS` used in API calls."""

    API_ITEM_CLS = "RowValue"
    """:class:`ApiItem`: Item class this list class holds."""

    @property
    def name(self):
        """Expose column name from :attr:`API_COLUMN`.

        Returns:
            :obj:`str`

        """
        return self.API_COLUMN.name

    @property
    def result_type(self):
        """Expose column result type from :attr:`API_COLUMN`.

        Returns:
            :obj:`str`

        """
        return self.API_COLUMN.result_type

    @property
    def hash(self):
        """Expose column sensor hash from :attr:`API_COLUMN`.

        Returns:
            :obj:`str`

        """
        return self.API_COLUMN.hash

    def get_values(self, attr="value", join=None):
        """Get all values from this row column.

        Args:
            attr (:obj:`str`, optional):
                Attribute to get values of.

                Defaults to: "value".
            join (:obj:`str`, optional):
                String to join multi value columns.

                Defaults to: None.

        Returns:
            :obj:`list` or :obj:`str`

        """
        values = []
        for row_value in self:
            value = getattr(row_value, attr)
            values.append("" if value is None else value)

        if join:
            values = join.join([format(v) for v in values])
        return values


class RowValue(ApiItem):
    """Manually defined API object."""

    API_NAME = "v"
    """:obj:`str`: Name of object used in API calls."""

    API_NAME_SRC = ""
    """:obj:`str`: Name of object in source file."""

    API_SIMPLE = {"h": "integer_types", "text": "simple_types"}
    """:obj:`dict`: Map of simple attributes to their types."""

    API_COMPLEX = {}
    """:obj:`dict`: Map of complex attributes to their types."""

    API_CONSTANTS = {}
    """:obj:`dict`: Map of constants to their values."""

    API_STR_ADD = ["hash", "value"]
    """:obj:`list` or :obj:`str`: Attributes to add to str formatting."""

    API_LIST_CLS = "RowColumn"
    """:class:`ApiList`: List class that holds this item class."""

    API_LIST_API_NAME = "row_column"
    """:obj:`str`: Name of :attr:`API_LIST_CLS` used in API calls."""

    @property
    def hash(self):
        """Expose simple attr "h" as hash.

        Returns:
            :obj:`str`

        """
        return self.h

    @property
    def value(self):
        """Expose simple attr "text" as value.

        Returns:
            :obj:`str`

        """
        return self.text



# -*- coding: utf-8 -*-
"""Adapter objects that serialize requests to the Tanium API."""
from __future__ import absolute_import
from __future__ import division
from __future__ import print_function
from __future__ import unicode_literals

import abc
import re
import six
import warnings
import xmltodict

from . import exceptions
from .. import api_models
from .. import results
from .. import utils


@six.add_metaclass(abc.ABCMeta)
class Adapter(object):
    """Abstract base class for all Adapters."""

    @abc.abstractproperty
    def api_objects(self):
        """Get the API objects container.

        Returns:
            :obj:`tantrum.api_objects.ApiObjects`

        """
        raise NotImplementedError  # pragma: no cover

    @abc.abstractproperty
    def api_client(self):
        """Get the API client.

        Returns:
            :obj:`tantrum.api_clients.ApiClient`

        """
        raise NotImplementedError  # pragma: no cover

    @abc.abstractproperty
    def http_client(self):
        """Get the HTTP client.

        Returns:
            :obj:`tantrum.http_client.HttpClient`

        """
        raise NotImplementedError  # pragma: no cover

    @abc.abstractproperty
    def auth_method(self):
        """Get the Auth Method.

        Returns:
            :obj:`tantrum.auth_methods.AuthMethod`

        """
        raise NotImplementedError  # pragma: no cover

    @abc.abstractproperty
    def result_cls(cls):
        """Get the result deserializer class.

        Returns:
            :class:`tantrum.results.Result`

        """
        raise NotImplementedError  # pragma: no cover

    @abc.abstractmethod
    def cmd_get(self, obj, **kwargs):
        """Send an API request to get an object.

        Args:
            obj (:obj:`tantrum.api_models.ApiModel`):
                API Object to use for request.

        Returns:
            :obj:`tantrum.results.Result`

        """
        raise NotImplementedError  # pragma: no cover

    @abc.abstractmethod
    def cmd_add(self, obj, **kwargs):
        """Send an API request to add an object.

        Args:
            obj (:obj:`tantrum.api_models.ApiModel`):
                API Object to use for request.

        Returns:
            :obj:`tantrum.results.Result`

        """
        raise NotImplementedError  # pragma: no cover

    @abc.abstractmethod
    def cmd_delete(self, obj, **kwargs):
        """Send an API request to delete an object.

        Args:
            obj (:obj:`tantrum.api_models.ApiModel`):
                API Object to use for request.

        Returns:
            :obj:`tantrum.results.Result`

        """
        raise NotImplementedError  # pragma: no cover

    @abc.abstractmethod
    def cmd_update(self, obj, **kwargs):
        """Send an API request to update an object.

        Args:
            obj (:obj:`tantrum.api_models.ApiModel`):
                API Object to use for request.

        Returns:
            :obj:`tantrum.results.Result`

        """
        raise NotImplementedError  # pragma: no cover

    @abc.abstractmethod
    def cmd_get_audit_logs(self, type, target, **kwargs):
        """Send an API request to get audit logs for an object.

        Args:
            type (:obj:`str`):
                Type of object to get audit logs of.
            target (:obj:`int`):
                ID of object type to get audit logs for.

        Returns:
            :obj:`tantrum.results.Result`

        """
        raise NotImplementedError  # pragma: no cover

    @abc.abstractmethod
    def cmd_get_client_count(self, **kwargs):
        """Send an API request to get the client count.

        Returns:
            :obj:`tantrum.results.Result`

        """
        raise NotImplementedError  # pragma: no cover

    @abc.abstractmethod
    def cmd_parse_question(self, text, **kwargs):
        """Send an API request to parse text.

        Args:
            text (:obj:`str`):
                Text to parse into question objects.

        Returns:
            :obj:`tantrum.results.Result`

        """
        raise NotImplementedError  # pragma: no cover

    @abc.abstractmethod
    def cmd_add_parsed_question(self, obj, **kwargs):
        """Send an API request to add a parsed question object.

        Args:
            obj (:obj:`tantrum.api_models.ApiModel`):
                API Object to use for request.

        Returns:
            :obj:`tantrum.results.Result`

        """
        raise NotImplementedError  # pragma: no cover

    @abc.abstractmethod
    def cmd_get_result_info(self, obj, **kwargs):
        """Send an API request to get result info for an object.

        Args:
            obj (:obj:`tantrum.api_models.ApiModel`):
                API Object to use for request.

        Returns:
            :obj:`tantrum.results.Result`

        """
        raise NotImplementedError  # pragma: no cover

    @abc.abstractmethod
    def cmd_get_result_data(self, obj, **kwargs):
        """Send an API request to get result data for an object.

        Args:
            obj (:obj:`tantrum.api_models.ApiModel`):
                API Object to use for request.

        Returns:
            :obj:`tantrum.results.Result`

        """
        raise NotImplementedError  # pragma: no cover

    @abc.abstractmethod
    def cmd_get_merged_result_data(self, objlist, **kwargs):
        """Send an API request to get merged result data for a list of objects.

        Args:
            objlist (:obj:`list`):
                List of API Objects to use for request.

        Returns:
            :obj:`tantrum.results.Result`

        """
        raise NotImplementedError  # pragma: no cover


class Soap(Adapter):
    """Tanium SOAP request adapter."""

    DEFAULT_OPTIONS = {}
    """:obj:`dict`: Default options to use in :meth:`build_options_from_kwargs`."""

    AUDIT_LOG_TYPES = [
        "authentication",
        "content_set",
        "content_set_role",
        "content_set_role_privilege",
        "dashboard",
        "dashboard_group",
        "group",
        "package_spec",
        "plugin_schedule",
        "saved_action",
        "saved_question",
        "sensor",
        "system_setting",
        "user",
        "user_group",
        "white_listed_url",
    ]
    """:obj:`list` of :obj:`str`: Valid types for :meth:`api_get_audit_logs`."""

    def __init__(self, api_client, api_objects, lvl="info"):
        """Constructor.

        Args:
            api_client (:obj:`tantrum.api_clients.ApiClient`):
                Client to use for sending API requests.
            api_objects (:obj:`tantrum.api_objects.ApiObjects`):
                API objects container to use for this adapter.
            lvl (:obj:`str`, optional):
                Logging level.

                Defaults to: "info".

        """
        self.log = utils.logs.get_obj_log(obj=self, lvl=lvl)
        """:obj:`logging.Logger`: Log."""

        self._api_objects = api_objects
        self._api_client = api_client

    def __str__(self):
        """Show object info.

        Returns:
            :obj:`str`

        """
        bits = [
            "api_objects={!r}".format(self.api_objects),
            "api_client={!r}".format(self.api_client),
            "http_client={!r}".format(self.http_client),
            "auth_method={!r}".format(self.auth_method),
        ]
        bits = "(\n  {},\n)".format(",\n  ".join(bits))
        cls = "{c.__module__}.{c.__name__}".format(c=self.__class__)
        return "{cls}{bits}".format(cls=cls, bits=bits)

    def __repr__(self):
        """Show object info.

        Returns:
            :obj:`str`

        """
        return self.__str__()

    @property
    def api_objects(self):
        """Get the API objects container.

        Returns:
            :obj:`tantrum.api_objects.ApiObjects`

        """
        return self._api_objects

    @property
    def api_client(self):
        """Get the API client.

        Returns:
            :obj:`tantrum.api_clients.ApiClient`

        """
        return self._api_client

    @property
    def http_client(self):
        """Get the HTTP client.

        Returns:
            :obj:`tantrum.http_client.HttpClient`

        """
        return self.api_client.http_client

    @property
    def auth_method(self):
        """Get the Auth Method.

        Returns:
            :obj:`tantrum.auth_methods.AuthMethod`

        """
        return self.api_client.auth_method

    @property
    def result_cls(cls):
        """Get the result deserializer class.

        Returns:
            :class:`tantrum.results.Result`

        """
        return results.Soap

    def build_options_from_kwargs(self, **kwargs):
        """Build an Options API object from kwargs and return the serialized form.

        Args:
            **kwargs:
                options_obj (:obj:`tantrum.api_models.ApiItem`):
                    A pre-established Options object.

                    Defaults to: new Options object from :attr:`api_objects`.
                rest of kwargs:
                    Set on Options object if key is an attr on object and attrs value
                    is None.

        Notes:
            Will set :attr:`DEFAULT_OPTIONS` as defaults to kwargs
            before applying values to Options object attributes.

        Returns:
            :obj:`dict`

        """
        default_options = getattr(self, "DEFAULT_OPTIONS", {}) or {}
        for k, v in default_options.items():
            kwargs.setdefault(k, v)
        opts = kwargs.pop("options_obj", self.api_objects.Options())
        check_object_type(obj=opts, types=(self.api_objects.Options,))
        for k in list(kwargs):
            if hasattr(opts, k) and getattr(opts, k, None) is None:
                setattr(opts, k, kwargs[k])
        return opts.serialize(wrap_name=False)

    def send(
        self,
        obj,
        cmd,
        body_re_limit=4000,
        ser_only_attrs=None,
        ser_exclude_attrs=None,
        ser_empty=None,
        ser_list_attrs=None,
        ser_wrap_name=None,
        ser_wrap_item_attr=None,
        verify=None,
        save_last=None,
        save_history=None,
        log_request=None,
        log_response=None,
        cause="",
        **kwargs
    ):
        """Build and send a SOAP API request.

        Args:
            obj (:obj:`tantrum.api_models.ApiModel`):
                ApiModel to serialize and send as part of request.
            cmd (:obj:`str`):
                SOAP Command to use in request.
            body_re_limit (:obj:`int`):
                Value to limit regex search of response body for <session> tag.

                Defaults to: 4000.
            ser_empty (:obj:`bool`):
                Include attributes that have a value of None when serializing.
                Uses False if None.

                Defaults to: None.
            ser_list_attrs (:obj:`bool`):
                Include simple attributes of :obj:`tantrum.api_models.ApiList`
                when serializing.
                Uses False if None.

                Defaults to: None.
            ser_exclude_attrs (:obj:`list` of :obj:`str`):
                Exclude these attributes when serializing.
                Uses [] if None.

                Defaults to: None.
            ser_only_attrs (:obj:`list` of :obj:`str`):
                Include only these attributes when serializing.
                Uses [] if None.

                Defaults to: None.
            ser_wrap_name (:obj:`bool`):
                Wrap the return in another dict whose key is set to the API name.
                Uses True if None.

                Defaults to: None.
            ser_wrap_item_attr (:obj:`bool`):
                Wrap list items in dict whose key is set to the API list
                item attribute.
                Uses True if None.

                Defaults to: None.
            verify (:obj:`bool` or :obj:`str`, optional):
                Enable SSL certification validation.
                If None uses :attr:`tantrum.http_client.HttpClient.verify`.

                Defaults to: None.
            save_last (:obj:`bool`, optional):
                Save last request to :attr:`HttpClient.last_request` and last response
                to :attr:`HttpClient.last_response`.
                If None uses :attr:`tantrum.http_client.HttpClient.save_last`.

                Defaults to: None.
            save_history (:obj:`bool`, optional):
                Append last response to :attr:`HttpClient.history`.
                If None uses :attr:`tantrum.http_client.HttpClient.save_history`.

                Defaults to: None.
            log_request (:obj:`bool`, optional):
                Log request details to debug level.
                If None uses :attr:`tantrum.http_client.HttpClient.log_request`.

                Defaults to: None.
            log_response (:obj:`bool`, optional):
                Log response details to debug level.
                If None uses :attr:`tantrum.http_client.HttpClient.log_response`.

                Defaults to: None.
            cause (:obj:`str`, optional):
                String to explain purpose of request.

                Defaults to: "".
            **kwargs:
                rest of kwargs:
                    Passed to :meth:`build_options_from_kwargs`.

        Raises:
            :obj:`exceptions.SessionNotFoundWarning`:
                If the <session> tag can not be found in the response body.

        Returns:
            :obj:`tantrum.results.Result`

        """
        ser_only_attrs = utils.tools.def_none(ser_only_attrs, [])
        ser_exclude_attrs = utils.tools.def_none(ser_exclude_attrs, [])
        ser_empty = utils.tools.def_none(ser_empty, False)
        ser_list_attrs = utils.tools.def_none(ser_list_attrs, False)
        ser_wrap_name = utils.tools.def_none(ser_wrap_name, True)
        ser_wrap_item_attr = utils.tools.def_none(ser_wrap_item_attr, True)

        if isinstance(obj, api_models.ApiModel):
            obj = obj.serialize(
                only_attrs=ser_only_attrs,
                exclude_attrs=ser_exclude_attrs,
                empty=ser_empty,
                list_attrs=ser_list_attrs,
                wrap_name=ser_wrap_name,
                wrap_item_attr=ser_wrap_item_attr,
            )

        opts = self.build_options_from_kwargs(**kwargs)

        request_dict = soap_envelope(cmd=cmd, obj=obj, opts=opts)
        request_body = serialize_xml(obj=request_dict)

        r = self.api_client(
            data=request_body,
            verify=verify,
            log_request=log_request,
            log_response=log_response,
            save_last=save_last,
            save_history=save_history,
            cause=cause,
        )

        try:
            auth_token = re_soap_tag(text=r.text, tag="session", limit=body_re_limit)
        except Exception:
            auth_token = ""

        if auth_token:
            self.api_client.auth_method.token = auth_token
        else:
            if body_re_limit:
                limit = "the first {}".format(body_re_limit)
            else:
                limit = "ALL"
            error = "XML tag 'session' not in {limit} characters of SOAP response body"
            error = error.format(limit=limit)
            warnings.warn(error, exceptions.SessionNotFoundWarning)
        return self.result_cls.from_response(
            api_objects=self.api_objects, response=r, lvl=self.log.level
        )

    def cmd_get(self, obj, **kwargs):
        """Send an API request to get an object.

        Args:
            obj (:obj:`tantrum.api_models.ApiModel`):
                API Object to use for request.
            **kwargs:
                rest of kwargs:
                    Passed to :meth:`send`.

        Returns:
            :obj:`tantrum.results.Result`

        """
        check_object_type(
            obj=obj, types=(self.api_objects.ApiItem, self.api_objects.ApiList)
        )
        kwargs["cmd"] = "GetObject"
        return self.send(obj=obj, **kwargs)

    def cmd_add(self, obj, **kwargs):
        """Send an API request to add an object.

        Args:
            obj (:obj:`tantrum.api_models.ApiModel`):
                API Object to use for request.
            **kwargs:
                rest of kwargs:
                    Passed to :meth:`send`.

        Returns:
            :obj:`tantrum.results.Result`

        """
        check_object_type(obj=obj, types=(self.api_objects.ApiItem,))
        kwargs["cmd"] = "AddObject"
        kwargs["ser_exclude_attrs"] = ["id"]
        return self.send(obj=obj, **kwargs)

    def cmd_delete(self, obj, **kwargs):
        """Send an API request to delete an object.

        Args:
            obj (:obj:`tantrum.api_models.ApiModel`):
                API Object to use for request.
            **kwargs:
                rest of kwargs:
                    Passed to :meth:`send`.

        Returns:
            :obj:`tantrum.results.Result`

        """
        check_object_type(obj=obj, types=(self.api_objects.ApiItem,))
        check_object_attrs(obj=obj, attrs=["id", "name"])
        kwargs["cmd"] = "DeleteObject"
        return self.send(obj=obj, **kwargs)

    def cmd_update(self, obj, **kwargs):
        """Send an API request to update an object.

        Args:
            obj (:obj:`tantrum.api_models.ApiModel`):
                API Object to use for request.
            **kwargs:
                rest of kwargs:
                    Passed to :meth:`send`.

        Returns:
            :obj:`tantrum.results.Result`

        """
        check_object_type(obj=obj, types=(self.api_objects.ApiItem,))
        check_object_attrs(obj=obj, attrs=["id", "name"])
        kwargs["cmd"] = "UpdateObject"
        return self.send(obj=obj, **kwargs)

    def cmd_get_audit_logs(self, type, target, **kwargs):
        """Send an API request to get audit logs for an object.

        Args:
            type (:obj:`str`):
                Type of object to get audit logs of.
            target (:obj:`int`):
                ID of object type to get audit logs for.
                SOAP allows target of 'None' to get all objects of `type`.
            **kwargs:
                count (:obj:`int`):
                    Limit number of audit logs returned to this.

                    Defaults to: 1.
                rest of kwargs:
                    Passed to :meth:`send`.

        Raises:
            :exc:`exceptions.InvalidTypeError`:
                If type is not one of :attr:`AUDIT_LOG_TYPES`.

        Returns:
            :obj:`tantrum.results.Result`

        """
        if type not in self.AUDIT_LOG_TYPES:
            error = "Invalid object type {ot} - MUST be one of {at}"
            error = error.format(at=self.AUDIT_LOG_TYPES, ot=type)
            raise exceptions.InvalidTypeError(error)

        obj = self.api_objects.AuditLog(type="{t}_audit".format(t=type), id=target)
        kwargs.setdefault("audit_history_size", kwargs.pop("count", 1))
        kwargs["cmd"] = "GetObject"
        return self.send(obj=obj, **kwargs)

    def cmd_get_client_count(self, **kwargs):
        """Send an API request to get the client count.

        Args:
            **kwargs:
                count (:obj:`int`):
                    Number of days to get client count for.

                    Defaults to: 30.
                rest of kwargs:
                    Passed to :meth:`send`.

        Returns:
            :obj:`tantrum.results.Result`

        """
        obj = {"client_count": kwargs.pop("count", 30)}
        kwargs["cmd"] = "GetObject"
        return self.send(obj=obj, **kwargs)

    def cmd_parse_question(self, text, **kwargs):
        """Send an API request to parse text.

        Args:
            text (:obj:`str`):
                Text to parse into question objects.
            **kwargs:
                rest of kwargs:
                    Passed to :meth:`send`.

        Returns:
            :obj:`tantrum.results.Result`

        """
        check_object_type(obj=text, types=six.string_types)
        obj = self.api_objects.ParseJob(question_text=text)
        kwargs["cmd"] = "AddObject"
        return self.send(obj=obj, **kwargs)

    def cmd_add_parsed_question(self, obj, **kwargs):
        """Send an API request to add a parsed question object.

        Args:
            obj (:obj:`tantrum.api_models.ApiModel`):
                API Object to use for request.
            **kwargs:
                rest of kwargs:
                    Passed to :meth:`send`.

        Returns:
            :obj:`tantrum.results.Result`

        """
        check_object_type(obj=obj, types=(self.api_objects.ParseResultGroup,))
        obj = obj.question
        kwargs["cmd"] = "AddObject"
        return self.send(obj=obj, **kwargs)

    def cmd_get_result_info(self, obj, **kwargs):
        """Send an API request to get result info for an object.

        Args:
            obj (:obj:`tantrum.api_models.ApiModel`):
                API Object to use for request.
            **kwargs:
                rest of kwargs:
                    Passed to :meth:`send`.

        Returns:
            :obj:`tantrum.results.Result`

        """
        types = (
            self.api_objects.Question,
            self.api_objects.SavedQuestion,
            self.api_objects.Action,
            self.api_objects.SavedAction,
        )
        check_object_type(obj=obj, types=types)
        check_object_attrs(obj=obj, attrs=["id", "name"])
        kwargs["cmd"] = "GetResultInfo"
        kwargs["ser_only_attrs"] = ["id", "name"]
        return self.send(obj=obj, **kwargs)

    def cmd_get_result_data(self, obj, **kwargs):
        """Send an API request to get result data for an object.

        Args:
            obj (:obj:`tantrum.api_models.ApiModel`):
                API Object to use for request.
            **kwargs:
                rest of kwargs:
                    Passed to :meth:`send`.

        Returns:
            :obj:`tantrum.results.Result`

        """
        types = (
            self.api_objects.Question,
            self.api_objects.SavedQuestion,
            self.api_objects.Action,
            self.api_objects.SavedAction,
        )
        check_object_type(obj=obj, types=types)
        check_object_attrs(obj=obj, attrs=["id", "name"])
        kwargs["cmd"] = "GetResultData"
        kwargs["ser_only_attrs"] = ["id", "name"]
        return self.send(obj=obj, **kwargs)

    def cmd_get_merged_result_data(self, objlist, **kwargs):
        """Send an API request to get merged result data for a list of objects.

        Args:
            objlist (:obj:`list`):
                List of API Objects to use for request.
            **kwargs:
                rest of kwargs:
                    Passed to :meth:`send`.

        Returns:
            :obj:`tantrum.results.Result`

        """
        list_types = (
            list,
            tuple,
            self.api_objects.SavedQuestionList,
            self.api_objects.QuestionList,
        )
        check_object_type(obj=objlist, types=list_types)

        item_types = (self.api_objects.Question, self.api_objects.SavedQuestion)
        pobjlist = [objlist] if isinstance(objlist, item_types) else objlist

        objs = {"question": [], "saved_question": []}
        only_attrs = ["id", "name", "index", "cache_row_id"]
        sargs = {
            "ser_wrap_name": False,
            "ser_wrap_item_attr": False,
            "ser_only_attrs": only_attrs,
        }
        idx = 0
        for pobj in pobjlist:
            sobjlist = [pobj] if isinstance(pobj, item_types) else pobj
            for sobj in sobjlist:
                check_object_type(obj=sobj, types=item_types)
                check_object_attrs(obj=sobj, attrs=["id", "name"])
                sobj.index = idx
                sobj.cache_row_id = getattr(sobj, "cache_row_id", None) or 0
                sobj_dict = sobj.serialize(**sargs)
                objs_target = sobj.API_NAME
                objs[objs_target].append(sobj_dict)
                idx += 1

        kwargs["cmd"] = "GetMergedResultData"
        return self.send(obj=objs, **kwargs)


def soap_envelope(cmd, obj, opts=None):
    """Construct a SOAP envelope with the request command, obj, and options.

    Args:
        cmd (:obj:`str`):
            Command to use for request.
        obj (:obj:`dict`):
            Object(s) to use for request.
        options (:obj:`dict`, optional):
            Options to use for request.

            Defaults to: None.

    Returns:
        :obj:`dict`

    """
    request = {"command": cmd, "object_list": obj, "options": opts}

    body = {
        "@xmlns:t": "urn:TaniumSOAP",
        "@xmlns:xsi": "http://www.w3.org/2001/XMLSchema-instance",
        "t:tanium_soap_request": request,
    }
    env = {
        "@soap:encodingStyle": "http://schemas.xmlsoap.org/soap/encoding/",
        "@xmlns:soap": "http://schemas.xmlsoap.org/soap/envelope/",
        "@xmlns:xsd": "http://www.w3.org/2001/XMLSchema",
        "soap:Body": body,
    }
    ret = {"soap:Envelope": env}
    return ret


def re_soap_tag(text, tag, limit=4000, pattern=r"<{t}>(.*?)</{t}>"):
    """Search for tag in text[:limit] using pattern.

    Args:
        text (:obj:`str`):
            Text to search for pattern.
        tag (:obj:`str`):
            Tag name to use in pattern as 't'.
        limit (:obj:`int`, optional):
            Length to limit text to when searching for pattern.

            Defaults to: 4000.
        pattern (:obj:`str`, optional):
            Pattern to use when searching for tag.

            Defaults to: r'<{e}>(.*?)</{e}>'

    Notes:
        Given text is 4 GB and pattern is expected at top of text:
            * if head is None and pattern not found: 131 seconds
            * if head is None and pattern found: 0 seconds
            * if head is 4000 and pattern not found: 0 seconds
            * if head is 4000 and pattern found: 0 seconds

    Returns:
        :obj:`str`

    """
    pattern_txt = pattern.format(t=tag)
    pattern_re = re.compile(pattern_txt, re.IGNORECASE | re.DOTALL)
    text_limit = text[:limit]
    match = pattern_re.search(text_limit)
    return match.group(1) if match else ""


def check_object_type(obj, types):
    """Check if an obj is an instance of types.

    Args:
        obj (:obj:`object`):
            Object to check against types.
        types (:obj:`tuple` of :obj:`type`):
            Types to check against obj.

    Raises:
        :exc:`exceptions.InvalidTypeError`:
            If type of obj is not on of types.

    """
    if not isinstance(obj, types):
        error = "Invalid object type {ot} - MUST be one of {at}"
        error = error.format(at=types, ot=type(obj))
        raise exceptions.InvalidTypeError(error)


def check_object_attrs(obj, attrs):
    """Check if any attributes of an obj are set.

    Args:
        obj (:obj:`tantrum.api_models.ApiModel`):
            API object to check.
        attrs (:obj:`list` of :obj:`str`):
            Attributes to check on obj.

    Raises:
        :exc:`exceptions.EmptyAttributeError`:
            If none of the attributes in attrs on obj are not set to None.

    """
    if isinstance(obj, api_models.ApiItem):
        if not any(getattr(obj, x, None) is not None for x in attrs):
            error = "No attributes in {a} defined on {o}"
            error = error.format(a=attrs, o=obj)
            raise exceptions.EmptyAttributeError(error)


def serialize_xml(obj, **kwargs):
    """Encode python object into an XML string.

    Args:
        obj (:obj:`object`):
            Python object to encode into a string.
        **kwargs:
            full_document (:obj:`bool`):
                Include xml stanza at top.

                Defaults to: True.
            pretty (:obj:`bool`):
                Indent the output doc.

                Defaults to: True.
            rest of kwargs:
                Passed to xmltodict.unparse.

    Returns:
        :obj:`str`

    """
    kwargs.setdefault("full_document", True)
    kwargs.setdefault("pretty", True)
    return xmltodict.unparse(obj, **kwargs)



#!/usr/bin/env python
# -*- coding: utf-8 -*-
"""Package setup."""
import os

from codecs import open

from setuptools import setup
from setuptools import find_packages

here = os.path.abspath(os.path.dirname(__file__))
version_path = os.path.join(here, "tantrum", "version.py")


about = {}
with open(version_path, "r", "utf-8") as f:
    x = f.readlines()
    contents = "\n".join(a for a in x if not a.startswith("#"))
    exec(contents, about)  # nosec


with open("README.md", "r", "utf-8") as f:
    readme = f.read()

packages = find_packages()

setup(
    name=about["__title__"],
    version=about["__version__"],
    description=about["__description__"],
    long_description=readme,
    long_description_content_type="text/markdown",
    author=about["__author__"],
    author_email=about["__author_email__"],
    url=about["__url__"],
    packages=packages,
    package_data={"": ["LICENSE"]},
    package_dir={"tantrum": "tantrum"},
    scripts=[],  # TODO(!)
    include_package_data=True,
    python_requires=">=2.7, !=3.0.*, !=3.1.*, !=3.2.*, !=3.3.*",
    install_requires=["requests[security,socks]", "six>=1.12.0", "xmltodict"],
    tests_require=[],
    license=about["__license__"],
    classifiers=[
        "Development Status :: 5 - Production/Stable",
        "Intended Audience :: Developers",
        "Natural Language :: English",
        "License :: OSI Approved :: MIT License",
        "Programming Language :: Python",
        "Programming Language :: Python :: 2",
        "Programming Language :: Python :: 2.7",
        "Programming Language :: Python :: 3",
        "Programming Language :: Python :: 3.7",
        "Programming Language :: Python :: Implementation :: CPython",
        "Programming Language :: Python :: Implementation :: PyPy",
    ],
)



#!/usr/bin/env python
# -*- coding: utf-8 -*-
"""Utility for downloading console.wsdl from a Tanium platform."""
from __future__ import absolute_import
from __future__ import division
from __future__ import print_function
from __future__ import unicode_literals

import argparse
import sys
import datetime

try:
    import pathlib
except Exception:
    import pathlib2 as pathlib

__version__ = "3.0.0"
TMPL_INIT = '''# -*- coding: utf-8 -*-
"""WSDL contents downloaded from Tanium platform server."""
from __future__ import absolute_import
from __future__ import division
from __future__ import print_function
from __future__ import unicode_literals

downloaded_on = "{date}"

from_version = "{version}"

from_url_path = "{url_path}"

data = """{data}

"""  # noqa
'''


def add_args(argparser, args):
    """Add arguments to argparser.

    Args:
        argparser (:obj:`argparse.ArgumentParser`):
            Object to add args to.
        args (:obj:`list` of :obj:`dict`):
            List of arguments to add to argparser

    Returns:
        :obj:`argparse.ArgumentParser`

    """
    required = argparser.add_argument_group("Required Arguments")
    optional = argparser.add_argument_group("Optional Arguments")

    for arg in args:
        arg_required = arg.get("required", False)
        arg_default = arg.get("default", argparse.SUPPRESS)
        arg_help = arg.get("help", "")
        arg_opts = arg.pop("opts") if "opts" in arg else []

        grp = required if arg_required else optional

        helps = []
        helps.append(arg_help)

        if arg_default is not argparse.SUPPRESS and not arg_required:
            helps.append("DEFAULT: {d!r}".format(d=arg_default))

        choices = arg.get("choices", [])
        if choices:
            helps.append("VALID VALUES: {c}".format(c=", ".join(choices)))

        arg["help"] = "\n".join(helps)
        grp.add_argument(*arg_opts, **arg)
    return argparser


def arg_parse(script_path):
    """Parse arguments from command line."""
    args = [
        {
            "opts": ["-u", "--url"],
            "dest": "url",
            "metavar": "https://{host}:{port}",
            "help": "URL to Tanium Platform Server",
            "required": True,
        },
        {
            "opts": ["-o", "--overwrite"],
            "dest": "overwrite",
            "default": False,
            "action": "store_true",
            "help": "Overwrite output objects files that already exist.",
            "required": False,
        },
        {
            "opts": ["-p", "--tantrum_path"],
            "dest": "tantrum_path",
            "default": script_path.parent.parent.parent.parent.parent,
            "help": "Path to tantrum package",
            "required": False,
        },
        {
            "opts": ["--output_dir"],
            "dest": "output_dir",
            "default": script_path.parent,
            "help": "Directory to create API version sub-directory with console.wsdl",
            "required": False,
        },
        {
            "opts": ["--version"],
            "help": "Version of this script.",
            "version": "%(prog)s {v}".format(v=__version__),
            "default": argparse.SUPPRESS,
            "action": "version",
        },
        {
            "opts": ["-h", "--help"],
            "help": "Show this help message and exit.",
            "default": argparse.SUPPRESS,
            "action": "help",
        },
    ]

    argparser = argparse.ArgumentParser(
        description=__doc__,
        add_help=False,
        formatter_class=argparse.RawTextHelpFormatter,
    )
    add_args(argparser=argparser, args=args)
    args, extra_args = argparser.parse_known_args()
    return args, extra_args


def check_dir(path, name):
    """Check that a directory exists.

    Args:
        path (:obj:`pathlib.Path` or :obj:`str`):
            Path to check.
        name (:obj:`str`):
            Reference name of path.

    Raises:
        :exc:`Exception`:
            If path does not exist.

    Returns:
        :obj:`pathlib.Path`

    """
    path = pathlib.Path(path).expanduser().absolute()

    if not path.is_dir():
        error = "Unable to find {name!r} path '{path!r}'"
        error = error.format(path=format(path), name=name)
        raise Exception(error)
    return path


def write_contents(path, contents, overwrite=False):
    """Write contents to path."""
    if path.is_file():
        if overwrite:
            m = "!! Overwrite is True - overwriting pre-existing file: {path!r}"
            m = m.format(path=format(path))
            print(m)
        else:
            m = "!! Overwrite is False, not overwriting pre-existing file: {path!r}"
            m = m.format(path=format(path))
            print(m)
            return
    path.write_text(contents)
    print("Wrote file: {}".format(path))


def main(url, output_dir, tantrum, overwrite=False):
    """Download a WSDL from a platform server to path named after version of server.

    Args:
        url (:obj:`str`):
            URL of platform server
        output_dir (:obj:`pathlib.Path` or :obj:`str`):
            Directory under which to create version directory.
        tantrum (:obj:`object`):
            tantrum3 module.
        overwrite (:obj:`bool`, optional):
            Over write WSDL file if exists.

            Defaults to: False.

    """
    output_dir = check_dir(path=output_dir, name="output")
    client = tantrum.http_client.HttpClient(url=url, verify=False)
    version = tantrum.api_clients.get_version(client)
    wsdl = tantrum.api_clients.get_wsdl(client)

    version_dir = output_dir / version.replace(".", "_")

    if not version_dir.is_dir():
        version_dir.mkdir()
        m = "Created directory {path!r}"
        m = m.format(path=format(version_dir))
        print(m)

    date = datetime.datetime.utcnow()
    url_path = tantrum.api_clients.WSDL_PATH

    wsdl_path = version_dir / "__init__.py"
    wsdl_contents = TMPL_INIT.format(
        data=wsdl, version=version, date=date, url_path=url_path
    )
    write_contents(path=wsdl_path, contents=wsdl_contents, overwrite=overwrite)

    m = "Done with version {version!r}, " "you need to manually populate 'soap.py'"
    m = m.format(version=version)
    print(m)


if __name__ == "__main__":
    script_path = pathlib.Path(sys.argv[0]).expanduser().absolute()
    args, _ = arg_parse(script_path)
    sys.path.insert(0, format(args.tantrum_path))
    import tantrum

    main(
        url=args.url,
        output_dir=args.output_dir,
        tantrum=tantrum,
        overwrite=args.overwrite,
    )
